"""
Summary.

    Commons Module -- Common Functionality

"""
import os
import sys
import json
import inspect
import logging
import platform
import subprocess
from shutil import which
from pathlib import Path
from xlines._version import __version__


logger = logging.getLogger(__version__)
logger.setLevel(logging.INFO)

try:
    from xlines.oscodes_unix import exit_codes
    os_type = 'Linux'
    user_home = os.getenv('HOME')
    splitchar = '/'                             # character for splitting paths (linux)
except Exception:
    from xlines.oscodes_win import exit_codes    # non-specific os-safe codes
    os_type = 'Windows'
    user_home = os.getenv('username')
    splitchar = '\\'                            # character for splitting paths (windows)


def get_os(detailed=False):
    """
    Summary:
        Retrieve local operating system environment characteristics
    Args:
        :user (str): USERNAME, only required when run on windows os
    Returns:
        TYPE: dict object containing key, value pairs describing
        os information
    """
    try:

        os_type = platform.system()

        if os_type == 'Linux':
            os_detail = platform.platform()
            distribution = platform.linux_distribution()[0]
            HOME = str(Path.home())
            username = os.getenv('USER')
        elif os_type == 'Windows':
            os_detail = platform.platform()
            username = os.getenv('username') or os.getenv('USER')
            HOME = 'C:\\Users\\' + username
        else:
            logger.warning('Unsupported OS. No information')
            os_type = 'Java'
            os_detail = 'unknown'
            HOME = os.getenv('HOME')
            username = os.getenv('USER')

    except OSError as e:
        raise e
    except Exception as e:
        logger.exception(
            '%s: problem determining local os environment %s' %
            (inspect.stack()[0][3], str(e))
            )
    if detailed and os_type == 'Linux':
        return {
                'os_type': os_type,
                'os_detail': os_detail,
                'linux_distribution': distribution,
                'username': username,
                'HOME': HOME
            }
    elif detailed:
        return {
                'os_type': os_type,
                'os_detail': os_detail,
                'username': username,
                'HOME': HOME
            }
    return {'os_type': os_type}


def import_file_object(filename):
    """
    Summary:
        Imports block filesystem object
    Args:
        :filename (str): block filesystem object
    Returns:
        dictionary obj (valid json file), file data object
    """
    try:
        handle = open(filename, 'r')
        file_obj = handle.read()
        dict_obj = json.loads(file_obj)

    except OSError as e:
        logger.critical(
            'import_file_object: %s error opening %s' % (str(e), str(filename))
        )
        raise e
    except ValueError:
        logger.info(
            '%s: import_file_object: %s not json. file object returned' %
            (inspect.stack()[0][3], str(filename))
        )
        return file_obj    # reg file, not valid json
    return dict_obj


def read_local_config(cfg):
    """ Parses local config file for override values

    Args:
        :local_file (str):  filename of local config file

    Returns:
        dict object of values contained in local config file
    """
    try:
        if os.path.exists(cfg):
            config = import_file_object(cfg)
            return config
        else:
            logger.warning(
                '%s: local config file (%s) not found, cannot be read' %
                (inspect.stack()[0][3], str(cfg)))
    except OSError as e:
        logger.warning(
            'import_file_object: %s error opening %s' % (str(e), str(cfg))
        )
    return {}


def os_parityPath(path):
    """
    Converts unix paths to correct windows equivalents.
    Unix native paths remain unchanged (no effect)
    """
    path = os.path.normpath(os.path.expanduser(path))
    if path.startswith('\\'):
        return 'C:' + path
    return path


def terminal_size(height=False):
    """
    Summary.

        Returns size of linux terminal rows, columns if
        called with height=True; else only width of terminal
        in columns is returned.

    Returns:
        columns (str, default) || rows, columns, TYPE: tuple

    """
    try:
        rows, columns = os.popen('stty size 2>/dev/null', 'r').read().split()
        if height:
            return rows, columns
        return columns
    except ValueError as e:
        if which('tput'):
            return subprocess.getoutput('tput cols')
        raise e


def user_home():
    """Returns os specific home dir for current user"""
    try:
        if platform.system() == 'Linux':
            return os.path.expanduser('~')

        elif platform.system() == 'Windows':
            username = os.getenv('username')
            return 'C:\\Users\\' + username

        elif platform.system() == 'Java':
            print('Unable to determine home dir, unsupported os type')
            sys.exit(1)
    except OSError as e:
        raise e



"""

Help Menu
    Help menu object containing body of help content.
    For printing with formatting

"""

from xlines.statics import PACKAGE
from xlines import Colors
from xlines.variables import *


c = Colors()

ACCENT = c.ORANGE               # orange accent highlight color
bdacct = c.ORANGE + c.BOLD      # bold orange
bdcy = c.CYAN + c.BOLD          # bold blue
lbrct = bbc + '[ ' + rst        # left bracket
rbrct = bbc + ' ]' + rst        # right bracket
vdiv = bbc + ' | ' + rst
tab = '\t'.expandtabs(24)

menu_title = '\n' + c.BOLD + tab + PACKAGE + rst + ' help contents'

synopsis_cmd = (
        rst + ACCENT + PACKAGE + rst + ' --sum <values> ' +
        lbrct + '--no-whitespace' + vdiv + '--multiprocess' + rbrct
    )

url_doc = c.URL + 'http://xlines.readthedocs.io' + rst
url_sc = c.URL + 'https://github.com/fstab50/xlines' + rst


footer = """
       --------------------------------------------------------
       |  """ + bbl + 'o' + rst + """  |  Filesystem object counted (""" + bcy + 'cyan' + rst + """)              |
       --------------------------------------------------------
       |  """ + bdacct + 'o' + rst + """  |  Line count above high ct threshold (""" + acct + 'orange' + rst + """)   |
       --------------------------------------------------------
       | """ + bwt + '->' + rst + """  |  Truncated (shortened) file path (""" + bwt + 'white' + rst + """)       |
       --------------------------------------------------------

  """


menu_body = menu_title + c.BOLD + """

  DESCRIPTION""" + rst + """

            Count lines of text: A utility for all code projects
            Source Code Repo:  """ + url_sc + """
    """ + c.BOLD + """
  SYNOPSIS""" + rst + """

        $ """ + synopsis_cmd + """

                        -s, --sum
                       [-e, --exclusions ]
                       [-c, --configure  ]
                       [-d, --debug  ]
                       [-h, --help   ]
                       [-m, --multiprocess  ]
                       [-n, --no-whitespace  ]
                       [-V, --version  ]
    """ + c.BOLD + """
  OPTIONS
        -s, --sum""" + rst + """ (string): Sum the counts of all lines contained
            in filesystem objects referenced in the sum parameter
    """ + c.BOLD + """
        -c, --configure""" + rst + """:  Configure runtime parameter via the cli
            menu. Change display format, color scheme, etc values
    """ + c.BOLD + """
        -d, --debug""" + rst + """:  Print out additional  debugging information
    """ + c.BOLD + """
        -e, --exclusions""" + rst + """:  Print out list of file type extensions
            and directories excluded from line count calculations
    """ + c.BOLD + """
        -h, --help""" + rst + """: Show this help message, symbol legend, & exit
    """ + c.BOLD + """
        -m, --multiprocess""" + rst + """:  Use multiple  cpu cores for counting
            lines of text in expansive filesystem directories
    """ + c.BOLD + """
        -w, --no-whitespace""" + rst + """:  Exclude lines containing whitespace
            from total line counts for all objects
    """ + c.BOLD + """
        -V, --version""" + rst + """:  Print package version  and copyright info
    """ + c.BOLD + """
  LEGEND""" + rst



"""
Summary:
    Module utilising unicode characters to create
    a help menu legend surrounded by a border

Module Funtions:
    - _map (dict lookup for legend content)
    - border_map (caller)

Caller:
    - cli.py module
"""

from xlines import Colors
from xlines.variables import c, acct, gn, bbl, bbc, bcy, bdcy, bwt, rst


c = Colors()

lbrct = bbc + '[ ' + rst        # left bracket
rbrct = bbc + ' ]' + rst        # right bracket
vdiv = bbc + ' | ' + rst
tab = '\t'.expandtabs(24)
tab6 = '\t'.expandtabs(8).encode('utf-8')


width = 55      # legend overall width

legend_content = [
        bdcy + '==' + rst + '  |  Filesystem object counted (' + bcy + 'cyan' + rst + ')',
        acct + '==' + rst + '  |  Line count above high ct threshold (' + gn + 'green' + rst + ')',
        bwt + '->' + rst + '  |  Truncated (shortened) file path (' + bwt + 'white' + rst + ')'
    ]

legend_width = 57
legend_bar_top = ('_' * legend_width) + '\n'
legend_bar_bottom = ('_' * legend_width)

res = [
    ('┌' + '─' * width + '┐').encode('utf-8')
]

fallback = """
         """ + bbl + 'o' + rst + """  |  Filesystem object counted (""" + bcy + 'cyan' + rst + """)
       ----------------------------------------------------------
         """ + acct + 'o' + rst + """  |  Line count above high ct threshold (""" + acct + 'green' + rst + """)
       ----------------------------------------------------------
        """ + bwt + '->' + rst + """  |  Truncated (shortened) file path (""" + bwt + 'white' + rst + """)
"""


def _map(index, content):
    return {
        0: ('│  ' + content + (' ' * int(width - 43)) + '  │').encode('utf-8'),
        1: ('│  ' + content + (' ' * int(width - 53)) + '  │').encode('utf-8'),
        2: ('│  ' + content + (' ' * int(width - 50)) + '  │').encode('utf-8')
    }.get(index)


def border_map(text_list=legend_content):
    """
        Render help menu legend using dict lookup module function

    Returns:
        decoded utf-8 characters printed to stdout

    """
    for index, s in enumerate(text_list):
        res.append(_map(index, s))

    res.append(('└' + '─' * width + '┘').encode('utf-8'))

    try:

        return [print((tab6 + x).decode('utf-8')) for x in res]

    except UnicodeEncodeError:
        # if problems handling unicode encoding
        text_list.insert(0, legend_bar_top)
        text_list.insert(len(text_list), legend_bar_bottom)
        print('\t'.expandtabs(8) + text_list[0])
        [print('\t'.expandtabs(10) + x) for x in text_list[1:-1]][0]
        print('\t'.expandtabs(8) + text_list[-1])

    except Exception:
        # if any utf-8 handling host issues
        print(fallback)



import sys
import json
import inspect
from pygments import highlight, lexers, formatters
from xlines import logger


def export_json_object(dict_obj, filename=None, logging=True):
    """
    Summary:
        exports object to block filesystem object

    Args:
        :dict_obj (dict): dictionary object
        :filename (str):  name of file to be exported (optional)

    Returns:
        True | False Boolean export status

    """
    def is_tty():
        """
        Summary:
            Determines if output is displayed to the screen or redirected
        Returns:
            True if tty terminal | False is redirected, TYPE: bool
        """
        return sys.stdout.isatty()

    try:

        if filename:

            try:

                with open(filename, 'w') as handle:
                    handle.write(json.dumps(dict_obj, indent=4, sort_keys=True))
                    logger.info(
                        '%s: Wrote %s to local filesystem location' %
                        (inspect.stack()[0][3], filename))

                handle.close()

            except TypeError as e:
                logger.warning(
                    '%s: object in dict not serializable: %s' %
                    (inspect.stack()[0][3], str(e)))

        elif is_tty():

            json_str = json.dumps(dict_obj, indent=4, sort_keys=True)

            print(
                highlight(
                    json_str,
                    lexers.JsonLexer(),
                    formatters.TerminalFormatter()
                ).strip()
            )
            if logging:
                logger.info('%s: successful export to stdout' % inspect.stack()[0][3])
            return True

        else:
            # print output, but not to stdout; possibly commandline redirect
            print(json.dumps(dict_obj, indent=4, sort_keys=True))

    except OSError as e:
        logger.critical(
            '%s: export_file_object: error writing to %s to filesystem. Error: %s' %
            (inspect.stack()[0][3], filename, str(e)))
        return False
    if logging:
        logger.info('export_file_object: successful export to %s' % filename)
    return True



#!/usr/bin/env python3
"""

xlines, GPL v3 License

Copyright (c) 2018-2019 Blake Huber

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the 'Software'), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

"""
import os
import sys
import re
import json
import inspect
import argparse
import subprocess
from shutil import copy2 as copyfile
from shutil import which
from pathlib import Path
from xlines import about, Colors, logger
from xlines.usermessage import stdout_message
from xlines.statics import local_config
from xlines.help_menu import menu_body
from xlines.square import border_map
from xlines.mp import multiprocessing_main
from xlines.core import absolute_paths, linecount, locate_fileobjects
from xlines.core import remove_illegal, print_footer, print_header
from xlines.exclusions import ExcludedTypes
from xlines.configure import display_exclusions, main_menupage
from xlines.colormap import ColorMap
from xlines.variables import *

cm = ColorMap()

try:
    from xlines.oscodes_unix import exit_codes
    os_type = 'Linux'
    user_home = os.getenv('HOME')
    splitchar = '/'                                   # character for splitting paths (linux)

except Exception:
    from xlines.oscodes_win import exit_codes         # non-specific os-safe codes
    os_type = 'Windows'
    user_home = os.getenv('username')
    splitchar = '\\'                                  # character for splitting paths (windows)


# globals
container = []
module = os.path.basename(__file__)
iloc = os.path.abspath(os.path.dirname(__file__))     # installed location of modules


def sp_linecount(path, abspath, exclusions):
    """
        Single threaded (sequential processing) line count

    Return:
        valid filesystem paths, TYPE: list

    """
    try:

        if os.path.isfile(path):
            return remove_illegal([path], exclusions)

        elif os.path.isdir(path):
            d = locate_fileobjects(path, abspath)
            valid_paths = remove_illegal(d, exclusions)
            return valid_paths

    except UnicodeDecodeError:
        pass


def filter_args(kwarg_dict, *args):
    """
    Summary:
        arg, kwarg validity test

    Args:
        :kwarg_dict: kwargs passed in to calling method or func
        :args:  valid keywords for the caller

    Returns:
        True if kwargs are valid; else raise exception
    """
    # unpack if iterable passed in args - TBD (here)
    if kwarg_dict is not None:
        keys = [key for key in kwarg_dict]
        unknown_arg = list(filter(lambda x: x not in args, keys))
        if unknown_arg:
            raise KeyError(
                '%s: unknown parameter(s) provided [%s]' %
                (inspect.stack()[0][3], str(unknown_arg))
            )
    return True


def help_menu():
    """
    Displays help menu contents
    """
    print(menu_body)
    border_map()
    sys.stdout.write('\n')
    return


def main(**kwargs):
    """ Main """
    keywords = ('root', 'debug', 'update', 'remediate')
    if filter_args(kwargs, *keywords):
        root = kwargs.get('root', user_home)
        update = kwargs.get('update', False)
        remediate = kwargs.get('remediate', False)
        debug = kwargs.get('debug', False)
    return False


def longest_path(parameters, exclusions):
    """
        Traces all subdirectories of provided commandline paths
        using MaxWidth object

    Args:
        :parameters (list): list of all sys.argv parameters supplied with --sum
        :exclusions (ExcludedTypes object): types to exclude

    Returns:
        width (integer), number of characters in longest path
    """
    mp = MaxWidth()     # max path object
    abspath = absolute_paths(parameters)
    container = []
    max_width = 0

    for i in parameters:
        try:
            paths = sp_linecount(i, abspath, exclusions.types)
            width = mp.calc_maxpath(paths)
            max_width = width if (width > max_width) else max_width
            container.extend(paths)
        except TypeError:
            stdout_message(message='Provided path appears to be invalid', prefix='WARN')
            sys.exit(exit_codes['EX_OSFILE']['Code'])
    return max_width, container


class MaxWidth():
    def __init__(self):
        self.buffer = local_config['OUTPUT']['COUNT_COLUMN_WIDTH'] + BUFFER
        self.term_width = os.get_terminal_size().columns - self.buffer - BUFFER
        self.max_width = local_config['OUTPUT']['MIN_WIDTH']

    def calc_maxpath(self, path_list):
        for path in path_list:
            if len(path) > self.max_width:
                self.max_width = len(path)
            if len(path) > self.term_width:
                continue
        return self.max_width if (self.max_width < self.term_width) else self.term_width


def modules_location():
    """Filsystem location of Python3 modules"""
    return os.path.split(os.path.abspath(__file__))[0]


def options(parser, help_menu=False):
    """
    Summary:
        parse cli parameter options

    Returns:
        TYPE: argparse object, parser argument set

    """
    parser.add_argument("-e", "--exclusions", dest='exclusions', action='store_true', required=False)
    parser.add_argument("-C", "--configure", dest='configure', action='store_true', required=False)
    parser.add_argument("-d", "--debug", dest='debug', action='store_true', default=False, required=False)
    parser.add_argument("-h", "--help", dest='help', action='store_true', required=False)
    parser.add_argument("-m", "--multiprocess", dest='multiprocess', default=False, action='store_true', required=False)
    parser.add_argument("-s", "--sum", dest='sum', nargs='*', default=os.getcwd(), required=False)
    parser.add_argument("-n", "--no-whitespace", dest='whitespace', action='store_false', default=True, required=False)
    parser.add_argument("-V", "--version", dest='version', action='store_true', required=False)
    return parser.parse_known_args()


def package_version():
    """
    Prints package version and requisite PACKAGE info
    """
    print(about.about_object)
    sys.exit(exit_codes['EX_OK']['Code'])


def set_hicount_threshold():
    """
        Retrieves high line count threshold from local filesystem if exists

    Returns:
        high line count threshold || None

    """
    local_linecount_file = local_config['CONFIG']['HI_THRESHOLD_FILEPATH']
    reference_threshold = local_config['OUTPUT']['COUNT_HI_THRESHOLD']

    try:

        if os.path.exists(local_linecount_file):
            with open(local_linecount_file) as f1:
                threshold = f1.read().strip()
        else:
            # not exist; create threshold file on local filesystem
            with open(local_linecount_file, 'w') as f1:
                f1.write(str(reference_threshold) + '\n')
                threshold = reference_threshold
    except OSError:
        fx = inspect.stack()[0][3]
        logger.exception(f'{fx}: Problem reading local hicount threshold file. Abort')
        return None
    return int(threshold)


def precheck(user_exfiles, user_exdirs, debug):
    """
    Runtime Dependency Checks: postinstall artifacts, environment
    """
    def set_environment():
        lang = 'undefined'
        if os.getenv('LANG') is None:
            lang = '{}export LANG=en_US.UTF-8{}'.format(yl, rst)
        elif 'UTF-8' not in os.getenv('LANG'):
            lang = '{}export LANG=$LANG.UTF-8{}'.format(yl, rst)
        return lang

    _os_configdir = os.path.join(modules_location(), 'config')
    _os_ex_fname = os.path.join(_os_configdir, local_config['EXCLUSIONS']['EX_FILENAME'])
    _os_dir_fname = os.path.join(_os_configdir, local_config['EXCLUSIONS']['EX_DIR_FILENAME'])
    _config_dir = local_config['CONFIG']['CONFIG_DIR']
    _language = set_environment()
    _environment_setup = 'fail' if 'UTF-8' in _language else 'success'

    if debug:
        tab = '\t'.expandtabs(16)
        stdout_message(f'_os_configdir: {_os_configdir}: system py modules location', 'DBUG')
        stdout_message(f'_os_ex_fname: {_os_ex_fname}: system exclusions.list path', 'DBUG')
        stdout_message(f'_os_dir_fname: {_os_dir_fname}: system directories.list file path', 'DBUG')
        stdout_message(f'_configdir: {_config_dir}: user home config file location', 'DBUG')
        stdout_message(f'Environment setup status: {_environment_setup.upper()}')

        if _environment_setup.upper() == 'FAIL':
            _env = _environment_setup.upper()
            msg = f'Environment setting is {_env}. Add the following code in your .bashrc file'
            stdout_message('{}:  {}'.format(msg, _language))

    try:
        # check if exists; copy
        if not os.path.exists(_config_dir):
            os.makedirs(_config_dir)

        # cp system config file to user if user config files absent
        if os.path.exists(_os_ex_fname) and os.path.exists(_os_dir_fname):

            if not os.path.exists(user_exfiles):
                copyfile(_os_ex_fname, user_exfiles)

            if not os.path.exists(user_exdirs):
                copyfile(_os_dir_fname, user_exdirs)

    except OSError:
        fx = inspect.stack()[0][3]
        logger.exception('{}: Problem installing user config files. Exit'.format(fx))
        return False
    return True


def create_container(parameters):
    """
    Summary.

        Create container list of fs artifacts to count

    Returns:
        list (iter)

    """
    container = []
    if isinstance(parameters, list):
        container.extend(parameters)
    else:
        container.append(parameters)
    return container


class RelpathProcessing():
    """Corrects relative paths"""
    def __init__(self):
        self.pattern_hidden = re.compile('^.[a-z]+')                    # hidden file (.xyz)
        self.pattern_asci = re.compile('^[a-z]+', re.IGNORECASE)        # regular, standalone file

    def normalize(self, path):
        """Prepends correct filesystem syntax if pwd counted"""
        if self.pattern_hidden.match(path) or self.pattern_asci.match(path):
            return './' + path


def init_cli():
    ex_files = local_config['EXCLUSIONS']['EX_EXT_PATH']
    ex_dirs = local_config['EXCLUSIONS']['EX_DIR_PATH']

    # process commandline args
    parser = argparse.ArgumentParser(add_help=False)

    try:

        args, unknown = options(parser)

    except Exception as e:
        help_menu()
        stdout_message(str(e), 'ERROR')
        sys.exit(exit_codes['E_BADARG']['Code'])

    # validate configuration files
    if precheck(ex_files, ex_dirs, args.debug):
        _ct_threshold = set_hicount_threshold() or local_config['CONFIG']['COUNT_HI_THRESHOLD']

    if len(sys.argv) == 1 or args.help:
        help_menu()
        sys.exit(exit_codes['EX_OK']['Code'])

    elif args.version:
        package_version()

    elif args.exclusions:
        display_exclusions(ex_files, ex_dirs)

    elif args.configure:
        main_menupage(ex_files, ex_dirs)

    elif len(sys.argv) == 2 and (sys.argv[1] != '.'):
        help_menu()
        sys.exit(exit_codes['EX_OK']['Code'])

    elif args.sum:

        ex = ExcludedTypes(ex_path=str(Path.home()) + '/.config/xlines/exclusions.list')
        container = create_container(args.sum)
        abspath = absolute_paths(container)

        if args.debug:
            stdout_message(f'xlines command line option parameter detail', prefix='DEBUG')
            print('\targs.sum: {}'.format(args.sum))
            print('\n\tsys.argv contents:\n')
            for i in sys.argv:
                print('\t\to  {}'.format(i))
            print(f'\n\tcontainer is:\t{container}')
            print(f'\n\tobject "unknown" is:\t{unknown}')
            print('\tabspath bool is {}\n'.format(abspath))
            print('\tmultiprocess bool is {}\n'.format(args.multiprocess))

        if args.multiprocess:
            # --- run with concurrency --
            width, paths = longest_path(container, ex)
            multiprocessing_main(paths, width, _ct_threshold, args.whitespace, ex, args.debug)

        elif not args.multiprocess:


            io_fail = []
            tcount, tobjects = 0, 0
            width, paths = longest_path(container, ex)

            print_header(width)
            count_width = local_config['OUTPUT']['COUNT_COLUMN_WIDTH']

            #for i in container:

            for path in paths:

                try:

                    inc = linecount(path, args.whitespace)
                    highlight = acct if inc > _ct_threshold else cm.aqu
                    tcount += inc    # total line count
                    tobjects += 1    # increment total number of objects

                    # truncation
                    lpath, fname = os.path.split(path)

                    if (len(path) + BUFFER * 2) > width:
                        cutoff = (len(path) + BUFFER * 2) - width
                    else:
                        cutoff = 0

                    tab = '\t'.expandtabs(width - len(lpath) - len(fname) - count_width + BUFFER)

                    # with color codes added
                    if cutoff == 0:
                        lpath = text + lpath + rst
                    else:
                        lpath = text + os.path.split(path)[0][:len(lpath) - cutoff - BUFFER] + rst + arrow
                        tab = '\t'.expandtabs(width - len(lpath) - len(fname) + count_width + BUFFER + cut_corr)

                    tab4 = '\t'.expandtabs(4)
                    fname = highlight + fname + rst

                    # incremental count formatting
                    ct_format = acct if inc > _ct_threshold else bwt

                    # format tabular line totals with commas
                    output_str = f'{tab4}{lpath}{div}{fname}{tab}{ct_format}{"{:,}".format(inc):>10}{rst}'
                    print(output_str)

                    if args.debug:
                        print(tab4*2 + 'lpath is {}'.format(lpath))
                        print(tab4*2 + 'fname is {}\n'.format(fname))

                except Exception:
                    io_fail.append(path)
                    continue

            print_footer(tcount, tobjects, width)

            if args.debug:
                tab4 = '\t'.expandtabs(4)
                stdout_message(f'cli screen columns variable, width: {cm.bdwt}{width}{cm.rst}', prefix='DBUG')
                print('\n' + tab4 + 'Skipped file objects:\n' + tab4 + ('-' * (width + count_width)))
                if io_fail:
                    for file in io_fail:
                        print('\t{}'.format(file))   # Write this out to a file in /tmp for later viewing
                else:
                    print('\tNone')
                sys.stdout.write('\n')

            sys.exit(exit_codes['EX_OK']['Code'])

    else:
        stdout_message(
            'Dependency check fail %s' % json.dumps(args, indent=4),
            prefix='AUTH',
            severity='WARNING'
            )
        sys.exit(exit_codes['E_DEPENDENCY']['Code'])

    failure = """ : Check of runtime parameters failed for unknown reason.
    Please ensure you have both read and write access to local filesystem. """
    logger.warning(failure + 'Exit. Code: %s' % sys.exit(exit_codes['E_MISC']['Code']))
    print(failure)



"""
Summary.

    Configuration Module -- configure run time parameters & exclusions

"""

import os
import sys
import inspect
import logging
from time import sleep
from xlines.exclusions import ProcessExclusions
from xlines.usermessage import stdout_message
from xlines.common import terminal_size
from xlines.colormap import ColorMap
from xlines._version import __version__
from xlines.statics import PACKAGE, local_config, exit_keywords, affirm_keywords
from xlines.variables import *

logger = logging.getLogger(__version__)
logger.setLevel(logging.INFO)

cm = ColorMap()
default_width = 4

try:

    from xlines.oscodes_unix import exit_codes
    os_type = 'Linux'

except Exception:
    from xlines.oscodes_win import exit_codes    # non-specific os-safe codes
    os_type = 'Windows'


def clearscreen():
    os.system('cls' if os.name == 'nt' else 'clear')
    return True


def _init_screen(starting_row='', height_percent=0.25):
    """

        Initializes screen before configuration artifact display

    Args:
        :starting_row (bool): return starting row from the top
            in addition to columns
        :height_percent (float): Decimal percentage of screen height

    Returns:
        - width in columns for centering artifacts on cli terminal
          right to left, TYPE: int
        - returns starting row (from top, optional), TYPE: int

    """
    clearscreen()
    rows, columns = terminal_size(height=True)
    if starting_row:
        print('\n' * int(starting_row))
        return int(columns)
    print('\n' * int(int(rows) * height_percent))
    return int(columns), int(int(rows) * height_percent)


def section_header(section, width=80, tabspaces=14):
    """
        Prints section header title and function ("section") with border

    Returns:
        width of section header in chars, TYPE: int
    """
    if section.lower() in ('add', 'delete'):
        title = '{} File Type Exclusions Menu'.format(bdwt + section.title() + rst)
    else:
        title = 'high line count {} update'.format(bdwt + section + rst)

    bar = '________________________________________________________________'
    pattern_width = len(bar)
    offset = '\t'.expandtabs(int((width / 2) - (pattern_width / 2)))
    border = bbl
    tab = '\t'.expandtabs(tabspaces)
    tab4 = '\t'.expandtabs(4)
    print(offset + border + bar + '\n' + rst)
    print('{}{}{}'.format(offset, tab, title))
    print(offset + border + bar + '\n' + rst)
    return len(bar)


def display_exclusions(expath, exdirpath, offset_spaces=default_width):
    """
    Show list of all file type extensions which are excluded
    from line total calculations
    """
    tab = '\t'.expandtabs(15)
    offset = '\t'.expandtabs(offset_spaces)

    # numbering
    div = cm.bpl + ')' + rst
    adj = 10

    try:

        ex = ProcessExclusions(expath)
        exclusions = ex.exclusions

        stdout_message(message='File types excluded from line counts:', indent=offset_spaces + adj)

        for index, ext in enumerate(exclusions):
            print('{}{}{:>3}{}'.format(offset, tab, index + 1, div + '  ' + ext))

        sys.stdout.write('\n')
        return True

    except OSError as e:
        fx = inspect.stack()[0][3]
        stdout_message(message=f'{fx}: Error: {e}. ', prefix='WARN')
        return False


def condition_map(letter, expath, exdirpath, srow):
    return {
        'a': _configure_add,
        'b': _configure_remove,
        'c': _configure_hicount,
        'd': lambda x, y: sys.exit
    }.get(letter, lambda x, y: None)(expath, exdirpath, srow)


def main_menupage(expath, exdirpath):
    """
    Displays main configuration menu jump page and options
    """
    def menu():
        border = bbl
        icolor = bbl
        bar = '\n' + ('_' * 80) + '\n'

        pattern_width = len(bar)
        width, srow = _init_screen()
        offset = '\t'.expandtabs(int((width / 2) - (pattern_width / 2)))
        _menu = (border + bar + rst + '''\n\n
            ''' + bdwt + PACKAGE + rst + ''' configuration main menu:\n\n
                  ''' + icolor + 'a' + rst + ''')  Add file type to exclusion list\n\n
                  ''' + icolor + 'b' + rst + ''')  Remove file type from exclusion list\n\n
                  ''' + icolor + 'c' + rst + ''')  Set high line count threshold (''' + acct + 'highlight' + rst + ''' file objects)\n\n
                  ''' + icolor + 'd' + rst + ''')  quit\n
        ''' + border + bar + rst)
        for line in _menu.split('\n'):
            print('{}{}'.format(offset, line))
        return offset, srow

    loop = True
    tab1 = '\t'.expandtabs(1)

    while loop:
        offset, verticalstart = menu()
        answer = input('\n{}{}Choose operation [quit]: '.format(offset, tab1)).lower()
        sys.stdout.write('\n')

        if (not answer) or (answer.lower() in exit_keywords):
            return True

        elif answer in ['a', 'b', 'c']:
            condition_map(answer, expath, exdirpath, verticalstart)

        else:
            stdout_message(
                    message='You must provide a letter a, b, c, or d',
                    indent=16,
                    prefix='INFO'
                )
            sys.stdout.write('\n')


def _configure_add(expath, exdirpath, startpt):
    """
        Add exclusions and update runtime constants

    Returns:
        Success | Failure, TYPE: bool
    """
    add_list = None
    tab4 = '\t'.expandtabs(4)
    loop = True
    adj = 4
    vert_adj = 2

    try:

        ex = ProcessExclusions(expath)
        exclusions = ex.exclusions

        while loop:
            width = _init_screen(starting_row=startpt + vert_adj)
            pattern_width = section_header('add', width, tabspaces=16)
            offset_chars = int((width / 2) - (pattern_width / 2)) + adj
            offset = '\t'.expandtabs(offset_chars)
            display_exclusions(expath, exdirpath, offset_chars)

            # query user input for new exclusions
            msg = 'Enter file extension types separated by commas [done]: '
            offset_msg = '\t'.expandtabs(int((width / 2) - (pattern_width / 2) + adj * 2))
            response = input(f'{offset_msg}{msg}')

            if not response:
                loop = False
                sys.stdout.write('\n')
                return True
            else:
                add_list = [x.strip() for x in response.split(',')]

                # add new extensions to existing
                exclusions.extend([x if x.startswith('.') else '.' + x for x in add_list])

            # exit loop; write out new exclusions config file
            with open(expath, 'w') as f2:
                f2.writelines([x + '\n' for x in exclusions])

    except OSError:
        stdout_message(
            message='Unable to modify local config file located at {}'.format(expath),
            prefix='WARN')
        return False


def _configure_rewrite(expath, newlist):
    """
        Rewrite existing exclusion list on local filesystem with
        modified contents from _configure operation

    Return:
        Succsss || Failure, TYPE: bool
    """
    try:
        # write new exclusion list to local disk
        with open(expath, 'w') as f1:
            list(filter(lambda x: f1.write(x.strip() + '\n'), newlist))

    except OSError as e:
        fx = inspect.stack()[0][3]
        stdout_message(
            f'{fx}: Problem writing new file type exclusion list: {expath}: {e}',
            prefix='WARN')
        return False
    return True


def continue_operation(offset_spaces):
    offset = '\t'.expandtabs(offset_spaces)
    msg = '{}Continue?  [quit]'.format(offset)
    answer = input(msg)
    if not answer or answer in ('quit', 'q', 'no', 'No'):
        return False
    return True


def _configure_remove(expath, exdirpath, startpt):
    """
        Remove file type extension from exclusion list

    Return:
        Succsss || Failure, TYPE: bool
    """
    delay_seconds = 2
    tabspaces = 4
    tab4 = '\t'.expandtabs(tabspaces)
    loop = True
    adj = tabspaces * 2
    vert_adj = 2

    try:

        # open current file type exclusions
        with open(expath) as f1:
            f2 = [x.strip() for x in f1.readlines()]

        while loop:
            width = _init_screen(starting_row=startpt + vert_adj)
            pattern_width = section_header('delete', width, tabspaces=14)
            offset_chars = int((width / 2) - (pattern_width / 2)) + tabspaces
            offset = '\t'.expandtabs(offset_chars)
            display_exclusions(expath, exdirpath, offset_chars)

            answer = input(offset + tab4 + 'Pick the number of a file type to remove [done]: ')

            try:
                if not answer:
                    loop = False
                    sys.stdout.write('\n')
                    return True

                elif int(answer) in range(1, len(f2) + 1):
                    # correct for f2 list index
                    answer = int(answer) - 1
                    # remove entry selected by user
                    deprecated = f2[answer]
                    f2.pop(int(answer))

                    if not _configure_rewrite(expath, f2):
                        return False

                    # Acknowledge removal
                    if str(answer) in f2:
                        stdout_message(
                            message='Failure to remove {} - reason unknown'.format(f2[answer]),
                            indent=offset_chars + adj,
                            prefix='FAIL'
                        )
                        sleep(delay_seconds)

                else:
                    max_index = len(f2)
                    stdout_message(
                        message=f'You must pick a number between 1 and {max_index}',
                        prefix='WARN',
                        indent=offset_chars + adj
                    )
                    sleep(delay_seconds)
            except ValueError:
                continue

    except OSError:
        stdout_message(
            message='Unable to modify local config file located at {}'.format(expath),
            prefix='WARN')
        return False


def _configure_hicount(expath, exdirpath, startpt):
    """
        User update high line count threshold persisted on local filesystem

    Returns:
        Success || Failure, TYPE: bool

    """
    def _exit(loop_break):
        leave = input('Exit? [quit]')
        if not leave or 'q' in leave:
            loop_break = False
        return loop_break

    tab4 = '\t'.expandtabs(4)
    tab13 = '\t'.expandtabs(13)
    loop = True
    adj = 12
    vert_adj = 2
    local_linecount_file = local_config['CONFIG']['HI_THRESHOLD_FILEPATH']

    try:
        width = _init_screen(starting_row=startpt + vert_adj)
        pattern_width = section_header('threshold', width, tabspaces=14)
        offset_chars = int((width / 2) - (pattern_width / 2))
        offset = '\t'.expandtabs(offset_chars)

        while loop:

            if os.path.exists(local_linecount_file):
                with open(local_linecount_file) as f1:
                    threshold = int(f1.read().strip())

            stdout_message(
                    message='Current high line count threshold: {}{}{}'.format(bdwt, threshold, rst),
                    indent=offset_chars + adj
                )
            answer = input(f'{tab4}{offset}{tab13}Enter high line count threshold [{threshold}]: ')

            try:
                if not answer:
                    stdout_message(
                            f'High line count threshold remains {threshold}',
                            prefix='INFO',
                            indent=offset_chars + adj
                        )
                    loop = False
                    return mainmenu_return(offset)

                elif 'q' in answer:
                    loop = False
                    return mainmenu_return(offset)

                elif type(int(answer)) is int:
                    # rewrite threshold file on local filesystem
                    with open(local_linecount_file, 'w') as f1:
                        f1.write(str(answer) + '\n')
                    stdout_message(
                            message='high line count threshold set to {}'.format(answer),
                            prefix='ok',
                            indent=offset_chars + adj
                        )
                    loop = False
                    return mainmenu_return(offset)

                else:
                    stdout_message(
                            message='You must enter an integer number',
                            prefix='INFO',
                            indent=offset_chars + adj
                        )
            except ValueError:
                pass
    except OSError:
        fx = inspect.stack()[0][3]
        logger.exception(f'{fx}: Problem reading local hicount threshold file. Abort')
        return False
    return True


def mainmenu_return(offset):
    """Return control to configuration main menu"""
    tab = '\t'.expandtabs(17)
    answer = input(f'{offset}{tab}Return to main menu [enter]: ')
    if (not answer) or (answer.lower() in (affirmative_keywords + exit_keywords)):
        return True
    return False



"""
Summary:
    Copyright, legal plate for display with PACKAGE version information
Args:
    url_doc (str): http url location pointer to official PACKAGE documentation
    url_sc (str):  http url location pointer to PACKAGE source code
    current_year (int): the current calendar year (4 digit)
Returns:
    copyright, legal objects
"""

import sys
import datetime
from xlines.statics import PACKAGE, LICENSE
from xlines import Colors
from xlines import __version__


# url formatting
url_doc = Colors.URL + 'https://xlines.readthedocs.io' + Colors.RESET
url_sc = Colors.URL + 'https://github.com/fstab50/xlines' + Colors.RESET
url_lic = Colors.URL + 'http://xlines.readthedocs.io/en/latest/license.html' + Colors.RESET

# copyright range thru current calendar year
current_year = datetime.datetime.today().year
copyright_range = '2017-' + str(current_year)

# python version number header
python_version = sys.version.split(' ')[0]
python_header = 'python'.title() + Colors.RESET + ' ' + python_version

# formatted package header
package_name = Colors.BOLD + PACKAGE + Colors.RESET


# --- package about statement -------------------------------------------------


title_separator = (
    ('\t').expandtabs(4) +
    '__________________________________________________________________\n\n\n'
    )

package_header = (
    '\n\t\t' + Colors.DARK_BLUE + PACKAGE + Colors.RESET + ' version: ' + Colors.WHITE +
    Colors.BOLD +  __version__ + Colors.RESET + '  |  ' + python_header + '\n\n'
    )

copyright = Colors.LT2GRAY + """
    Copyright """ + copyright_range + """, Blake Huber.  This program distributed under
    """ + LICENSE + """.  This copyright notice must remain with derivative works.""" + Colors.RESET + """
    __________________________________________________________________
        """ + Colors.RESET

about_object = """
""" + title_separator + """

""" + package_header + """


    __________________________________________________________________
""" + copyright



"""
Project-level logging module

"""
import os
import sys
import inspect
import logging
import logging.handlers
from pathlib import Path
from xlines.statics import local_config
from xlines._version import __version__


syslog = logging.getLogger(__version__)
syslog.setLevel(logging.DEBUG)


def mode_assignment(arg):
    """
    Translates arg to enforce proper assignment
    """
    arg = arg.upper()
    stream_args = ('STREAM', 'CONSOLE', 'STDOUT')
    try:
        if arg in stream_args:
            return 'STREAM'
        else:
            return arg
    except Exception:
        return None


def logging_prep(mode):
    """
    Summary:
        prerequisites for log file generation
    Return:
        Success | Failure, TYPE: bool
    """
    try:
        if mode == 'FILE':

            log_path = local_config['LOGGING']['LOG_PATH']
            # path: path to log dir
            path, log_dirname = os.path.split(log_path)

            if not os.path.exists(path):
                os.makedirs(path)

            if not os.path.exists(log_path):
                Path(log_path).touch(mode=0o644, exist_ok=True)

    except OSError as e:
        syslog.exception(f'{inspect.stack()[0][3]}: Failure while seeding log file path: {e}')
        return False
    return True


def getLogger(*args, **kwargs):
    """
    Summary:
        custom format logger

    Args:
        mode (str):  The Logger module supprts the following log modes:

            - log to console / stdout. Log_mode = 'stream'
            - log to file
            - log to system logger (syslog)

    Returns:
        logger object | TYPE: logging
    """

    log_mode = local_config['LOGGING']['LOG_MODE']

    # log format - file
    file_format = '%(asctime)s - %(pathname)s - %(name)s - [%(levelname)s]: %(message)s'

    # log format - stream
    stream_format = '%(pathname)s - %(name)s - [%(levelname)s]: %(message)s'

    # log format - syslog
    syslog_format = '- %(pathname)s - %(name)s - [%(levelname)s]: %(message)s'
    # set facility for syslog:
    if local_config['LOGGING']['SYSLOG_FILE']:
        syslog_facility = 'local7'
    else:
        syslog_facility = 'user'

    # all formats
    asctime_format = "%Y-%m-%d %H:%M:%S"

    # objects
    logger = logging.getLogger(*args, **kwargs)
    logger.propagate = False

    try:
        if not logger.handlers:
            # branch on output format, default to stream
            if mode_assignment(log_mode) == 'FILE':

                # file handler
                if logging_prep(mode_assignment(log_mode)):

                    f_handler = logging.FileHandler(local_config['LOGGING']['LOG_PATH'])
                    f_formatter = logging.Formatter(file_format, asctime_format)
                    f_handler.setFormatter(f_formatter)
                    logger.addHandler(f_handler)
                    logger.setLevel(logging.DEBUG)

                else:
                    syslog.warning(f'{inspect.stack()[0][3]}: Log preparation fail - exit')
                    sys.exit(1)

            elif mode_assignment(log_mode) == 'STREAM':
                # stream handlers
                s_handler = logging.StreamHandler()
                s_formatter = logging.Formatter(stream_format)
                s_handler.setFormatter(s_formatter)
                logger.addHandler(s_handler)
                logger.setLevel(logging.DEBUG)

            elif mode_assignment(log_mode) == 'SYSLOG':
                sys_handler = logging.handlers.SysLogHandler(address='/dev/log', facility=syslog_facility)
                sys_formatter = logging.Formatter(syslog_format)
                sys_handler.setFormatter(sys_formatter)
                logger.addHandler(sys_handler)
                logger.setLevel(logging.DEBUG)

            else:
                syslog.warning(
                    '%s: [WARN]: log_mode value of (%s) unrecognized - not supported' %
                    (inspect.stack()[0][3], str(log_mode))
                    )
                ex = Exception(
                    '%s: Unsupported mode indicated by log_mode value: %s' %
                    (inspect.stack()[0][3], str(log_mode))
                    )
                raise ex
    except OSError as e:
        raise e
    return logger



"""
Summary.

    Multiprocessing module for line count

Module Functions:


"""
import os
import multiprocessing
import inspect
from xlines.usermessage import stdout_message
from xlines import Colors
from xlines.core import BUFFER, acct, bwt, text, rst, arrow, div
from xlines.core import linecount, print_header, print_footer
from xlines.export import export_json_object
from xlines import local_config, logger
from xlines.variables import *


def cpu_cores(logical=True):
    """
        Finds number of physical and logical cores on machine
        Defaults to logical cores unless logical set to False

    Returns:
        # of logical cores (int) || physical cores (int)

    """
    physical_cores = multiprocessing.cpu_count()
    return len(os.sched_getaffinity(0)) if logical else physical_cores


def longest_path_mp(object_list):
    """
        Multiprocessing version of longest_path() module function. Sorts a
        list of dict by a key value (path) in order of descending length. Allows
        easy extraction of the longest path in position [0] dict in the list.

    Args:
        :object_list (list):  Format:
            [
                {
                    'path': '/var/lib/dpkg/info/xtrans-dev.list',
                    'count': 531
                },
                {
                    'path': '/var/lib/dpkg/transitive.list',
                    'count': 249
                },
            ]

    Returns:
        longest path (int)

    """
    s = sorted(object_list, key=lambda x: len(x['path']), reverse=True)
    return len(s[0]['path'])


def mp_linecount(path_list, exclusions, no_whitespace):
    """Multiprocessing line count"""
    for path in path_list:
        try:
            if os.path.isfile(path):

                path_value = os.path.abspath(path) if path.startswith('/') else ('./' + os.path.relpath(path))

                q.put(
                        {
                            'path': path_value,
                            'count': linecount(path, no_whitespace)
                        }
                    )

            elif os.path.isdir(path):
                for p in [os.path.join(path, x) for x in os.listdir(path)]:
                    q.put({'path': p, 'count': linecount(p, no_whitespace)})

        except UnicodeDecodeError:
            continue


def print_results(object_list, _ct_threshold, width):
    """
        Outputs paths and filesystem objects to which line counts
        were calculated.  Single process operation combines output
        from multiprocessing count by recombining lists of dict
        processed separately by each cpu core

    Args:
        :object_list (list):  Format:

        .. code: json

            [
                {
                    'path': '/var/lib/dpkg/info/xtrans-dev.list',
                    'count': 531
                },
                {
                    'path': '/var/lib/dpkg/transitive.list',
                    'count': 249
                },
            ]

        :_ct_threshold (int): path rec highlight color if object linecount
            is at or above this threshold
        :width (int): width in characters of the output pattern

    Returns:
        True | False, TYPE: bool

    """
    tcount, tobjects = 0, 0
    io_fail = []

    print_header(width)
    count_width = local_config['OUTPUT']['COUNT_COLUMN_WIDTH']

    for object_dict in sorted(object_list, key=lambda x: x['path']):

        try:
            path = object_dict['path']
            inc = object_dict['count']
            highlight = acct if inc > _ct_threshold else Colors.AQUA
            tcount += inc    # total line count
            tobjects += 1    # increment total number of objects

            # truncation
            lpath, fname = os.path.split(path)

            if (len(path) + BUFFER * 2) > width:
                cutoff = (len(path) + BUFFER * 2) - width
            else:
                cutoff = 0

            tab = '\t'.expandtabs(width - len(lpath) - len(fname) - count_width + BUFFER)
            tab4 = '\t'.expandtabs(4)

            # with color codes added
            if cutoff == 0:
                lpath = text + lpath + rst
            else:
                lpath = text + os.path.split(path)[0][:len(lpath) - cutoff - BUFFER] + rst + arrow
                tab = '\t'.expandtabs(width - len(lpath) - len(fname) + count_width + BUFFER + cut_corr)

            fname = highlight + fname + rst

            # incremental count formatting
            ct_format = acct if inc > _ct_threshold else bwt

            output_str = f'{tab4}{lpath}{div}{fname}{tab}{ct_format}{"{:,}".format(inc):>10}{rst}'
            print(output_str)
        except Exception:
            io_fail.append(path)
            continue

    print_footer(tcount, tobjects, width)
    return True


def split_list(mlist, n):
    """
    Summary.

        splits a list into equal parts as allowed, given n segments

    Args:
        :mlist (list):  a single list containing multiple elements
        :n (int):  Number of segments in which to split the list

    Returns:
        generator object

    """
    k, m = divmod(len(mlist), n)
    return (mlist[i * k + min(i, m):(i + 1) * k + min(i + 1, m)] for i in range(n))


def get_varname(var):
    callers_local_vars = inspect.currentframe().f_back.f_locals.items()
    return [var_name for var_name, var_val in callers_local_vars if var_val is var]


def multiprocessing_main(valid_paths, max_width, _threshold, wspace, exclusions, debug):
    """
        Execute Operations using concurrency (multi-process) model

    Args:
        :valid_paths (list): list of filesystem paths (str) filtered for binary,
            other uncountable objects
        :max_width (int): width of output pattern, sans line count total column
        :_threshold (int): high line count threshold value (highlighted objects)
        :wspace (bool): when True, omit whitespace lines from count (DEFAULT:  False)
        :exclusions (ex object): instance of ExcludedTypes
        :debug (boot): debug flag

    """
    def debug_messages(flag, paths):
        if flag:
            stdout_message('Objects contained in container directories:', prefix='DEBUG')
            for i in paths:
                print(i)

    def queue_generator(q, p):
        """Generator which offloads a queue before full"""
        while p.is_alive():
            p.join(timeout=1)
            while not q.empty():
                yield q.get(block=False)

    global q
    q = multiprocessing.Queue()
    processes, results = [], []
    debug_messages(debug, valid_paths)

    # maximum cores is 4 due to i/o contention single drive systems
    cores = 4 if cpu_cores() >= 4 else cpu_cores()
    equal_lists = [sorted(x) for x in split_list(valid_paths, cores)]

    for i in equal_lists:
        t = multiprocessing.Process(target=mp_linecount, args=(i, exclusions.types, wspace))
        processes.append(t)
        t.start()

        results.extend([x for x in queue_generator(q, t)])

        if debug:
            print('Completed: list {}'.format(get_varname(i)))    # show progress

    print_results(results, _threshold, max_width)

    if debug:
        export_json_object(results, logging=False)
        stdout_message(message='Num of objects: {}'.format(len(results)))
    return 0



"""
Standard OS Module Exit Codes
    - See https://docs.python.org/3.6/library/os.html#process-management

Module Attributes:
    - exit_codes (dict): exit error codes for Unix, Linux

"""

import os


# --- exit codes for Unix, Linux, Linux-based operating systems ----------------


exit_codes = {
    'EX_OK': {
        'Code': 0,
        'Reason': 'No error occurred'
    },
    'E_DEPENDENCY': {
        'Code': 1,
        'Reason': 'Missing required dependency'
    },
    'E_DIR': {
        'Code': 2,
        'Reason': 'Failure to create log dir, log file'
    },
    'E_ENVIRONMENT': {
        'Code': 3,
        'Reason': 'Incorrect shell, language interpreter, or operating system'
    },
    'EX_AWSCLI': {
        'Code': 4,
        'Reason': 'Value could not be determined from local awscli configuration'
    },
    'EX_NOPERM': {
        'Code': os.EX_NOPERM,
        'Reason': 'IAM user or role permissions do not allow this action'
    },
    'E_AUTHFAIL': {
        'Code': 5,
        'Reason': 'Authentication Fail'
    },
    'E_BADPROFILE': {
        'Code': 6,
        'Reason': 'Local profile variable not set or incorrect'
    },
    'E_USER_CANCEL': {
        'Code': 7,
        'Reason': 'User abort'
    },
    'E_BADARG': {
        'Code': 8,
        'Reason': 'Bad input parameter'
    },
    'E_EXPIRED_CREDS': {
        'Code': 9,
        'Reason': 'Credentials expired or otherwise no longer valid'
    },
    'E_MISC': {
        'Code': 9,
        'Reason': 'Unknown Error'
    },
    'EX_NOUSER': {
        'Code': os.EX_NOUSER,
        'Reason': 'specified user does not exist'
    },
    'EX_CONFIG': {
        'Code': os.EX_CONFIG,
        'Reason': 'Configuration or config parameter error'
    },
    'EX_CREATE_FAIL': {
        'Code': 21,
        'Reason': 'Keyset failed to create. Possible Permissions issue'
    },
    'EX_DELETE_FAIL': {
        'Code': 22,
        'Reason': 'Keyset failed to delete.  Possible Permissions issue'
    },
    'EX_DATAERR': {
        'Code': os.EX_DATAERR,
        'Reason': 'Input data incorrect'
    },
    'EX_NOINPUT': {
        'Code': os.EX_NOINPUT,
        'Reason': 'Input file does not exist or not readable.'
    },
    'EX_UNAVAILABLE': {
        'Code': os.EX_UNAVAILABLE,
        'Reason': 'Required service or dependency unavailable.'
    },
    'EX_PROTOCOL': {
        'Code': os.EX_PROTOCOL,
        'Reason': 'Protocol exchange was illegal, invalid, or not understood.'
    },
    'EX_OSERR': {
        'Code': os.EX_OSERR,
        'Reason': 'Operating system error.'
    },
    'EX_OSFILE': {
        'Code': os.EX_OSFILE,
        'Reason': 'System file does not exist, or could not be opened'
    },
    'EX_IOERR': {
        'Code': os.EX_IOERR,
        'Reason': 'Exit code that means that an error occurred while doing I/O on some file.'
    },
    'EX_NOHOST': {
        'Code': os.EX_NOHOST,
        'Reason': 'Network host does not exist or not found'
    },
    'EX_SOFTWARE': {
        'Code': os.EX_SOFTWARE,
        'Reason': 'Internal software error detected.'
    },
    'EX_CANTCREAT': {
        'Code': os.EX_CANTCREAT,
        'Reason': 'User specified output file could not be created.'
    }
}



"""
Summary:
    ANSI color and formatting code class
    See: http://www.lihaoyi.com/post/BuildyourownCommandLinewithANSIescapecodes.html#256-colors
    VERSION:  1.13
Args:
    None

Returns:
    ansi codes

Raises:
    None.  AttributeError if no code match returns the reset ansi codes
"""


class Colors():
    """
    Class attributes provide different format variations
    """
    # forground colors
    AQUA = '\u001b[38;5;14m'
    PURPLE = '\033[95m'
    CYAN = '\033[96m'
    DARK_CYAN = '\033[36m'
    BLUE = '\033[94m'
    DARK_BLUE = '\033[38;5;95;38;5;24m'
    PURE_BLUE = '\u001b[38;5;27m'
    GREEN = '\033[92m'
    DARK_GREEN = '\u001b[38;5;2m'
    YELLOW = '\033[93m'
    RED = '\033[91m'
    ORANGE = '\033[38;5;95;38;5;214m'
    GOLD2 = '\033[38;5;95;38;5;136m'                    # yellow-orange, lighter than GOLD3
    GOLD3 = '\033[38;5;95;38;5;178m'                    # yellow-orange
    WHITE = '\033[37m'
    WHITE_GRAY = '\033[38;5;95;38;5;250m'                # white-gray
    LT1GRAY = '\033[38;5;95;38;5;245m'                  # light gray
    LT2GRAY = '\033[38;5;95;38;5;246m'
    LT3GRAY = '\u001b[38;5;249m'
    DARK_GRAY1 = '\033[90m'
    DARK_GRAY2 = '\033[38;5;95;38;5;8m'                  # darkest gray

    # bright colors
    BRIGHT_BLUE = '\033[38;5;51m'
    BRIGHT_CYAN = '\033[38;5;36m'
    BRIGHT_GREEN = '\033[38;5;95;38;5;46m'
    BRIGHT_PURPLE = '\033[38;5;68m'
    BRIGHT_RED = '\u001b[31;1m'
    BRIGHT_YELLOW = '\033[38;5;11m'
    BRIGHT_YELLOW2 = '\033[38;5;95;38;5;226m'
    BRIGHT_YELLOWGREEN = '\033[38;5;95;38;5;155m'
    BRIGHT_WHITE = '\033[38;5;15m'

    # background colors
    BKGND_BLACK = '\u001b[0m'
    BKGND_GREY = '\u001b[40m'
    BKGND_BLUE = '\u001b[44m'
    BKGND_CYAN = '\u001b[46m'
    BKGND_GREEN = '\u001b[42m'
    BKGND_MAGENTA = '\u001b[45m'
    BKGND_RED = '\u001b[41m'
    BKGND_WHITE = '\u001b[47m'
    BKGND_WHITE_BOLD = '\u001b[47;1m'
    BKGND_YELLOW = '\u001b[43m'

    # background colors; bright
    BKGND_BRIGHT_BLACK = '\u001b[40;1m'
    BKGND_BRIGHT_BLUE = '\u001b[44;1m'
    BKGND_BRIGHT_CYAN = '\u001b[46;1m'
    BKGND_BRIGHT_GREEN = '\u001b[42;1m'
    BKGND_BRIGHT_GRAY = '\u001b[38;1m'
    BKGND_BRIGHT_MAGENTA = '\u001b[45;1m'
    BKGND_BRIGHT_RED = '\u001b[41;1m'
    BKGND_BRIGHT_WHITE = '\u001b[47;1m'
    BKGND_BRIGHT_YELLOW = '\u001b[43;1m'

    # formats
    BOLD = '\033[1m'
    UNBOLD = '\033[22m'
    UNDERLINE = '\033[4m'
    ITALIC = '\e[3m'
    END = '\033[0m'
    REVERSE = '\033[;7m'
    RESET = '\033[0;0m'
    RESET_ALT = '\u001b[0m'

    # special formats
    URL = UNDERLINE + CYAN
    TITLE = UNDERLINE + BOLD

    #except AttributeError as e:
    #    logger.info('Ansi color code not found (%s), returning reset code' % str(e))



"""
Curses-menu
"""
import sys
import logging
from xlines import __version__
from xlines.configure import display_exclusions

try:
    from cursesmenu import CursesMenu, SelectionMenu
    from cursesmenu.items import FunctionItem, SubmenuItem, CommandItem
except Exception as e:
    print('Errors when importing module cursesmenu: %s' % str(e))


logger = logging.getLogger(__version__)
logger.setLevel(logging.INFO)


def submenu_backup(menu_obj):
    submenu = CursesMenu("Access Key Backup")
    q1_item = FunctionItem("Do you want to retain a copy of newly created access keys?", input, ["yes"])
    q2_item = FunctionItem("Enter the directory where a backup copy of the access keys should be stored", input, ["~/Backup"])
    submenu.append_item(q1_item)
    submenu.append_item(q2_item)
    submenu_item = SubmenuItem("Show a submenu", submenu, menu=menu_obj)
    return submenu_item


def submenu_logging(menu_obj):
    submenu = CursesMenu("Logging Configuration")
    q1_item = FunctionItem("Do you want to turn enable logging?", input, ['yes'])
    q2_item = FunctionItem("If yes, do you want logging to stdout or a log file?", input, ['FILE'])
    submenu.append_item(q1_item)
    submenu.append_item(q2_item)
    submenu_item = SubmenuItem("Show a submenu", submenu, menu=menu_obj)
    return submenu_item


def selection_menu_example():
    bool_list = ['~/logs', '~/Backup']
    submenu2 = SelectionMenu(bool_list, "What directory location?")
    submenu2.show()
    submenu2.join()
    selection = submenu2.selected_option
    print('\nselection is: %s\n' % str(selection))


def main():
    """ Builds main menu, branches to submenus """

    # parent menu
    menu = CursesMenu("Local Configuration Menu", "keyup Project")

    try:
        submenu_backup = submenu_backup(menu)
        submenu_logging = submenu_logging(menu)

        # assemble main menu
        menu.append_item(submenu_backup)
        menu.append_item(submenu_logging)

        menu.show()
        user_selection = menu.selected_option
    except Exception as e:
        print('Unknown Exception: %s. Exit' % str(e))
        return False
    return True


if __name__ == '__main__':
    menu = CursesMenu("Local Configuration Menu", "xlines Project")
    # backup
    submenu = CursesMenu("Exclusions", 'Do you want to add a new file type exclusion?')
    q1_item1 = FunctionItem("Yes", input, ["yes"])
    q1_item2 = FunctionItem("No", print('Exit'), ["no"])
    submenu.append_item(q1_item1)
    submenu.append_item(q1_item2)
    submenu_exclusions = SubmenuItem("Configure Exclusion List", submenu, menu=menu)
    menu.append_item(submenu_exclusions)
    if q1_item1 == "yes":
        display_exclusions()
        submenu = CursesMenu("Access Key Backup", 'Enter the directory where a backup copy of the access keys should be stored')
        q2_item1 = FunctionItem("~/Backup/", input, ["~/Backup"])
        submenu.append(q2_item1)
        submenu.show()

    # logging
    submenu2 = CursesMenu("Logging Configuration", "Do you want to enable logging?")
    q2_item1 = FunctionItem("Yes", selection_menu_example(), ['yes'])
    q2_item2 = FunctionItem("No", input, ['No'])
    print('\nsubmenu_selection is: %s\n' % str(selection_menu_example()))
    submenu2.append_item(q2_item1)
    submenu2.append_item(q2_item2)

    submenu_logging = SubmenuItem("Configure Logging", submenu2, menu=menu)
    menu.append_item(submenu_logging)
    menu.show()
    user_selection = menu.selected_option
    print('\nuser_selection is: %s\n' % str(user_selection))
    sys.exit(0)



"""
Python3 Module

Facility for printing messages to stdout
    - Python3 only
    - Developed & Tested under Python3.6

Message Prefixes:

     Status                 Color
  -----------------      ------------------
    - INFO                  Cyan (DEFAULT)
    - ERROR                 Red
    - WARN                  Orange
    - OK                    Green
    - <user defined>        Cyan

Severity
"""
import inspect
from xlines.colors import Colors


# prefix handling
critical_status = ('ERROR', 'FAIL', 'WTF', 'STOP', 'HALT', 'EXIT', 'F*CK')
warning_status = (
    'WARN', 'WARNING', 'CAUTION', 'SLOW', 'DBUG', 'DEBUG', 'AUTH')


def stdout_message(message, prefix='INFO', quiet=False, multiline=False, indent=4, severity=''):
    """
    Summary:
        Prints message to cli stdout while indicating type and severity

    Args:
        :message (str): text characters to be printed to stdout
        :prefix (str):  4-letter string message type identifier.
        :quiet (bool):  Flag to suppress all output
        :multiline (bool): indicates multiline message; removes blank lines
            on either side of printed message
        :indent (int): left justified number of spaces to indent before
            printing message ouput
        :severity (str): header status msg determines color instead of prefix

    .. code-block:: python

        # Examples:

            - INFO (default)
            - ERROR (error, problem occurred)
            - WARN (warning)
            - NOTE (important to know)

    Returns:
        TYPE: bool, Success (printed) | Failure (no output)
    """
    prefix = prefix.upper()
    tabspaces = int(indent)

    if quiet:

        return True

    else:

        try:

            if prefix in critical_status or severity.upper() == 'CRITICAL':
                header = (Colors.YELLOW + '\t[ ' + Colors.RED + prefix +
                          Colors.YELLOW + ' ]' + Colors.RESET + ': ')

            elif prefix in warning_status or severity.upper() == 'WARNING':
                header = (Colors.YELLOW + '\t[ ' + Colors.ORANGE + prefix +
                          Colors.YELLOW + ' ]' + Colors.RESET + ': ')

            elif prefix in ('OK') or severity.upper() == 'OK':
                header = (Colors.YELLOW + '\t[  ' + Colors.BOLD + Colors.GREEN + prefix +
                          Colors.YELLOW + '  ]' + Colors.RESET + ': ')

            elif prefix in ('DONE', 'GOOD') or severity.upper() == 'OK':
                header = (Colors.YELLOW + '\t[ ' + Colors.BOLD + Colors.GREEN + prefix +
                          Colors.YELLOW + ' ]' + Colors.RESET + ': ')

            else:    # default color scheme
                header = (Colors.YELLOW + '\t[ ' + Colors.DARK_CYAN + prefix +
                          Colors.YELLOW + ' ]' + Colors.RESET + ': ')

            if multiline:
                print(header.expandtabs(tabspaces) + str(message))
            else:
                print('\n' + header.expandtabs(tabspaces) + str(message) + '\n')

        except Exception as e:
            print(f'{inspect.stack()[0][3]}: Problem sending msg to stdout: {e}')
            return False

    return True



"""
Summary.

    Core Logic Module -- Common Functionality

"""
import os
import sys
import re
import inspect
import logging
from shutil import which
from xlines.colors import Colors
from xlines.statics import local_config
from xlines._version import __version__
from xlines.variables import *

logger = logging.getLogger(__version__)
logger.setLevel(logging.INFO)

try:
    from xlines.oscodes_unix import exit_codes
    os_type = 'Linux'
    user_home = os.getenv('HOME')
    splitchar = '/'                             # character for splitting paths (linux)
    acct = Colors.ORANGE
    text = Colors.BRIGHT_PURPLE
    TITLE = Colors.WHITE + Colors.BOLD
except Exception:
    from xlines.oscodes_win import exit_codes    # non-specific os-safe codes
    os_type = 'Windows'
    user_home = os.getenv('username')
    splitchar = '\\'                            # character for splitting paths (windows)
    acct = Colors.CYAN
    text = Colors.LT2GRAY
    TITLE = Colors.WHITE + Colors.BOLD


def absolute_paths(path_list):
    """
        Determines if filesystem paths are absolute or relative

    Args:
        :path_list (list):
            [
                '/usr/local/bin/python3',
                '/home/user/myfile.txt',
                '../myfile.txt'
            ]

    Returns:
        True (absolute paths) || False (relative  paths), TYPE: bool

    """
    prefix = '/'
    if any(i.startswith(prefix) for i in path_list):
        return True
    return False


def is_binary_external(filepath):
    try:
        f = open(filepath, 'rb').read(1024)
        textchars = bytearray({7, 8, 9, 10, 12, 13, 27} | set(range(0x20, 0x100)) - {0x7f})
        fx = lambda bytes: bool(bytes.translate(None, textchars))
    except Exception:
        return True
    return fx(f)


def linecount(path, whitespace=True):
    if whitespace:
        return len(open(path).readlines())
    return len(list(filter(lambda x: x != '\n', open(path).readlines())))


def remove_duplicates(duplicates):
    """
    Summary.

        Module function utilsing a generator to remove duplicates
        from large scale lists with minimal resource use

    Args:
        duplicates (list): contains repeated elements

    Returns:
        list object (iter)
    """
    uniq = []

    def dedup(d):
        for element in d:
            if element not in uniq:
                uniq.append(element)
                yield element
    return [x for x in dedup(duplicates)]


def remove_illegal(d, illegal):
    """
        Removes excluded file types

    Args:
        :d (list): list of filesystem paths ending with a file object
        :illegal (list):  list of file type extensions for to be excluded

    Returns:
        legal filesystem paths (str)
    """
    def parse_list(path):
        """Reads in list from file object"""
        with open(path) as f1:
            return [x.strip() for x in f1.readlines()]

    def is_binary(filepath):
        try:
            f = open(filepath, 'rb').read(1024)
            textchars = bytearray({7, 8, 9, 10, 12, 13, 27} | set(range(0x20, 0x100)) - {0x7f})
            fx = lambda bytes: bool(bytes.translate(None, textchars))
            return fx(f)
        except Exception:
            return True

    bad = []

    try:
        illegal_dirs = parse_list(local_config['EXCLUSIONS']['EX_DIR_PATH'])
    except KeyError:
        illegal_dirs = ['pycache', 'venv']

    # filter for illegal or binary file object
    for fpath in d:

        fobject = os.path.split(fpath)[1]

        # filter for illegal dirs first, then files, then binary
        if list(filter(lambda x: x in fpath, illegal_dirs)):
            bad.append(fpath)

        elif ('.' in fobject) and ('.' + fobject.split('.')[1] in illegal):
            bad.append(fpath)

        if is_binary(fpath):
            bad.append(fpath)

    return sorted(list(set(d) - set(bad)))


def locate_fileobjects(origin, abspath=True):
    """
    Summary.

        - Walks local fs directories identifying all git repositories

    Args:
        - origin (str): filesystem directory location
        - abspath (bool): return absolute paths relative to current cursor position

    Returns:
        - paths, TYPE: list
        - Format:

         .. code-block:: json

                [
                    '/cloud-custodian/tools/c7n_mailer/c7n_mailer/utils_email.py',
                    '/cloud-custodian/tools/c7n_mailer/c7n_mailer/slack_delivery.py',
                    '/cloud-custodian/tools/c7n_mailer/c7n_mailer/datadog_delivery.py',
                    '/cloud-custodian/tools/c7n_sentry/setup.py',
                    '/cloud-custodian/tools/c7n_sentry/test_sentry.py',
                    '/cloud-custodian/tools/c7n_kube/setup.py',
                    '...
                ]

    """
    def relpath_normalize(path):
        """
        Prepends correct relative filesystem syntax if analyzed pwd
        """
        if pattern_hidden.match(path) or pattern_asci.match(path):
            return './' + path
        elif path.startswith('..'):
            return path

    pattern_hidden = re.compile('^.[a-z]+')                    # hidden file (.xyz)
    pattern_asci = re.compile('^[a-z]+', re.IGNORECASE)        # standalone, regular file
    fobjects = []

    if os.path.isfile(origin):
        return [origin]

    for root, dirs, files in os.walk(origin):
        for file in [f for f in files if '.git' not in root]:
            try:

                if abspath:
                    # absolute paths (default)
                    _path = os.path.abspath(os.path.join(root, file))
                else:
                    # relative paths (optional)
                    _path = os.path.relpath(os.path.join(root, file))
                    _path = relpath_normalize(_path)

                fobjects.append(_path)

            except OSError:
                logger.exception(
                    '%s: Read error while examining local filesystem path (%s)' %
                    (inspect.stack()[0][3], _path)
                )
                continue
    return remove_duplicates(fobjects)


def print_header(w):
    total_width = w + local_config['OUTPUT']['COUNT_COLUMN_WIDTH'] + 1
    header_lhs = 'object'
    header_rhs = 'line count'
    tab = '\t'.expandtabs(total_width - len(header_lhs) - len(header_rhs))
    tab4 = '\t'.expandtabs(4)
    print(tab4 + (horiz * (total_width)))
    print(f'{tab4}{header_lhs}{tab}{header_rhs}')
    print(tab4 + (horiz * (total_width)))


def print_footer(total, object_count, w):
    """
    Print total number of objects and cumulative total line count
    """
    total_width = w + local_config['OUTPUT']['COUNT_COLUMN_WIDTH'] + 1

    # add commas
    total_lines = '{:,}'.format(object_count)

    # calc dimensions; no color codes
    msg = 'Total ({} objects):'.format(total_lines)
    tab = '\t'.expandtabs(total_width - len(msg) - len(str(total)) - 1)

    # redefine with color codes added
    msg = f'Total ({title + "{:,}".format(object_count) + rst} objects):'
    tab4 = '\t'.expandtabs(4)

    # divider pattern
    print(tab4 + (horiz * (total_width)))

    # ending summary stats line
    print(f'{tab4}{msg}{tab}{highlight + "{:,}".format(total) + rst:>10}' + '\n')



"""
Summary.

    Exclusions Module -- Display, Update, and Modification
    of objects to be excluded from line count totals

"""

import os
import inspect
from shutil import copy2 as copyfile
from xlines.statics import local_config
from xlines import logger

module = os.path.basename(__file__)
library_location = os.path.split(module)[0]
types_fname = local_config['EXCLUSIONS']['EX_FILENAME']
dirs_fname = local_config['EXCLUSIONS']['EX_DIR_FILENAME']
excluded_types = os.path.join(library_location, types_fname)
excluded_dirs = os.path.join(library_location, dirs_fname)
config_location = local_config['CONFIG']['CONFIG_DIR']


class ExcludedTypes():
    """
        Class for processing file type exclusions (exclusions.list)
        File located in local configuration directory (~/.config/xlines)
    """
    def __init__(self, ex_path, ex_container=[]):
        """
        Args:
            ex_path (str): path to exclusions.list file
            ex_container (list): in memory list of all file extensions
                                 to be excluded from line counts
        """
        self.types = ex_container
        if not self.types:
            self.types.extend(self.parse_exclusions(ex_path))

    def excluded(self, path):
        return (True if list(filter(lambda x: x in path, self.types)) else False)

    def parse_exclusions(self, path):
        """
        Parse persistent fs location store for file extensions to exclude
        """
        try:
            return [x.strip() for x in open(path).readlines()]
        except OSError:
            exclusions = local_config['EXCLUSIONS']['EX_EXT_PATH']
            fx = inspect.stack()[0][3]
            logger.exception('{}: Problem parsing file type exclusions ({})'.format(fx, exclusions))
            return []


class ProcessExclusions():
    """
        Class definition for reading excluded file and directory
        types from a configuration file if exists.

    Use:
        >>> pex = ProcessExclusions()
        >>> pex.get_exclusions('/path/to/configuration/file')
        >>> ['.docx', '.pptx', '.png', '.tiff']

    Returns:
        exclusions (list) |  [] (empty) and creates new config
        file if not found

    """
    def __init__(self, path):
        """
        Args:
            :path (str): filesystem path to file object for parsing

        """
        self.path = path
        self.basedir = os.path.dirname(path)
        self.exclusions = self.process_exclusions(path)

    def process_exclusions(self, path=None):
        """
        Read and parse local config file if exists
        """
        valid_path = self.path if path is None else path
        if os.path.exists(valid_path):
            try:
                with open(valid_path) as f1:
                    return [x.strip() for x in f1.readlines()]
            except OSError:
                logger.info(f'path provided does not yet exist -- creating')
        return self._touch(self.path) if path is None else self._touch(path)

    def _touch(self, path):
        """
        If not exist, create new filesystem objects to hold
            excluded fs patterns
        """
        if not os.path.exists(config_location):
            os.makedirs(config_location)
        if not os.path.exists(path):
            self._create_ex_generic(path)
        return self.process_exclusions(os.path.join(config_location, types_fname))

    def _create_ex_generic(self, path):
        """Create excluded file types file via reference"""
        _fname = os.path.split(path)[1]
        copyfile(excluded_types, os.path.join(config_location, _fname))

    def _create_ex_types(self, path):
        """Create excluded file types file via reference"""
        copyfile(excluded_types, os.path.join(config_location, types_fname))

    def _create_ex_dirfile(self, path):
        """Create excluded file types file via reference"""
        copyfile(excluded_dirs, os.path.join(config_location, dirs_fname))



from xlines.colors import Colors
from xlines._version import __version__ as version
from xlines import logd
from xlines.statics import local_config


__author__ = 'Blake Huber'
__version__ = version
__email__ = "blakeca00@gmail.com"


# global logger
logd.local_config = local_config
logger = logd.getLogger(__version__)



"""
    Static assignments for universally used colors and variables

"""
import os
from xlines.statics import local_config
from xlines.colors import Colors


c = Colors()

try:

    from xlines.oscodes_unix import exit_codes
    os_type = 'Linux'
    user_home = os.getenv('HOME')
    splitchar = '/'                                     # character for splitting paths (linux)

    # special colors - linux
    acct = c.BOLD + c.BRIGHT_GREEN
    text = c.BRIGHT_PURPLE
    TITLE = c.WHITE + c.BOLD

except Exception:
    from xlines.oscodes_win import exit_codes           # non-specific os-safe codes
    os_type = 'Windows'
    username = os.getenv('username')
    splitchar = '\\'                                    # character for splitting paths (windows)
    user_home = 'Users' + splitchar + username
    # special colors - windows
    acct = c.CYAN
    text = c.LT2GRAY
    TITLE = c.WHITE + c.BOLD


# universal colors
rd = c.RED + c.BOLD
yl = c.YELLOW + c.BOLD
fs = c.GOLD3
bd = c.BOLD
gn = c.BRIGHT_GREEN
title = c.BRIGHT_WHITE + c.BOLD
bcy = c.BRIGHT_CYAN
bdcy = c.CYAN + c.BOLD          # bold blue
bbc = bd + c.BRIGHT_CYAN
bbl = bd + c.BRIGHT_BLUE
highlight = bd + c.BRIGHT_BLUE
frame = gn + bd
btext = text + c.BOLD
bwt = c.BRIGHT_WHITE
bdwt = c.BOLD + c.BRIGHT_WHITE
ub = c.UNBOLD
rst = c.RESET

# globals
expath = local_config['EXCLUSIONS']['EX_EXT_PATH']      # file path for file type exclusions file
exdirpath = local_config['EXCLUSIONS']['EX_DIR_PATH']   # file path for directory exclusions file
config_dir = local_config['CONFIG']['CONFIG_PATH']      # path to localhost configuration directory
div = text + bd + '/' + rst                             # path / file object divider character
div_len = 2
horiz = text + '-' + rst
arrow = bwt + '-> ' + rst
BUFFER = local_config['OUTPUT']['BUFFER']
cut_corr = 18                                           # correction factor, cut paths



"""
Standard OS Module Exit Codes
    - See https://docs.python.org/3.6/library/os.html#process-management

Module Attributes:
    - exit_codes (dict):  exist error codes for Microsoft Windows
"""

import os


# --- exit codes for Microsoft Windows operating systems -----------------------


exit_codes = {
    'EX_OK': {
        'Code': 0,
        'Reason': 'No error occurred'
    },
    'E_DEPENDENCY': {
        'Code': 1,
        'Reason': 'Missing required dependency'
    },
    'E_DIR': {
        'Code': 2,
        'Reason': 'Failure to create log dir, log file'
    },
    'E_ENVIRONMENT': {
        'Code': 3,
        'Reason': 'Incorrect shell, language interpreter, or operating system'
    },
    'EX_AWSCLI': {
        'Code': 4,
        'Reason': 'Value could not be determined from local awscli configuration'
    },
    'EX_NOPERM': {
        'Code': 77,
        'Reason': 'IAM user or role permissions do not allow this action'
    },
    'E_AUTHFAIL': {
        'Code': 5,
        'Reason': 'Authentication Fail'
    },
    'E_BADPROFILE': {
        'Code': 6,
        'Reason': 'Local profile variable not set or incorrect'
    },
    'E_USER_CANCEL': {
        'Code': 7,
        'Reason': 'User abort'
    },
    'E_BADARG': {
        'Code': 8,
        'Reason': 'Bad input parameter'
    },
    'E_EXPIRED_CREDS': {
        'Code': 9,
        'Reason': 'Credentials expired or otherwise no longer valid'
    },
    'E_MISC': {
        'Code': 9,
        'Reason': 'Unknown Error'
    },
    'EX_NOUSER': {
        'Code': 67,
        'Reason': 'specified user does not exist'
    },
    'EX_CONFIG': {
        'Code': 78,
        'Reason': 'Configuration or config parameter error'
    },
    'EX_CREATE_FAIL': {
        'Code': 21,
        'Reason': 'Keyset failed to create. Possible Permissions issue'
    },
    'EX_DELETE_FAIL': {
        'Code': 22,
        'Reason': 'Keyset failed to delete.  Possible Permissions issue'
    }
}
